//
//  exampleMNIST.cpp
//  MANITEST
//
//  Created by Alhussein Fawzi on 24.07.15.
//  Copyright (c) 2015 Alhussein Fawzi. All rights reserved.
//

#include "exampleMNIST.h"



int ReverseInt (int i)
{
    unsigned char ch1, ch2, ch3, ch4;
    ch1 = i & 255;
    ch2 = (i >> 8) & 255;
    ch3 = (i >> 16) & 255;
    ch4 = (i >> 24) & 255;
    return((int) ch1 << 24) + ((int)ch2 << 16) + ((int)ch3 << 8) + ch4;
}


void read_Mnist(std::string filename, std::vector<cv::Mat> &vec){
    std::ifstream file (filename, std::ios::binary);
    if (file.is_open())
    {
        int magic_number = 0;
        int number_of_images = 0;
        int n_rows = 0;
        int n_cols = 0;
        file.read((char*) &magic_number, sizeof(magic_number));
        magic_number = ReverseInt(magic_number);
        file.read((char*) &number_of_images,sizeof(number_of_images));
        number_of_images = ReverseInt(number_of_images);
        file.read((char*) &n_rows, sizeof(n_rows));
        n_rows = ReverseInt(n_rows);
        file.read((char*) &n_cols, sizeof(n_cols));
        n_cols = ReverseInt(n_cols);
        
        for(int i = 0; i < number_of_images; ++i)
        {
            // cv::Mat tp = Mat::zeros(n_rows, n_cols, CV_8UC1);
            cv::Mat tp = cv::Mat::zeros(n_rows, n_cols, CV_32F);
            for(int r = 0; r < n_rows; ++r)
            {
                for(int c = 0; c < n_cols; ++c)
                {
                    unsigned char temp = 0;
                    file.read((char*) &temp, sizeof(temp));
                    // tp.at<uchar>(r, c) = (int) temp;
                    tp.at<float>(r, c) = (float) temp;
                    
                }
            }
            vec.push_back(tp);
        }
    }
}

void read_Mnist32(std::string filename, std::vector<cv::Mat> &vec){
    std::ifstream file (filename, std::ios::binary);
    if (file.is_open())
    {
        int magic_number = 0;
        int number_of_images = 0;
        int n_rows = 0;
        int n_cols = 0;
        
        file.read((char*) &magic_number, sizeof(magic_number));
        magic_number = ReverseInt(magic_number);
        file.read((char*) &number_of_images,sizeof(number_of_images));
        number_of_images = ReverseInt(number_of_images);
        file.read((char*) &n_rows, sizeof(n_rows));
        n_rows = ReverseInt(n_rows);
        file.read((char*) &n_cols, sizeof(n_cols));
        n_cols = ReverseInt(n_cols);
        
        for(int i = 0; i < number_of_images; ++i)
        {
            // cv::Mat tp = Mat::zeros(n_rows, n_cols, CV_8UC1);
            cv::Mat tp = cv::Mat::zeros(32, 32, CV_32F);
            for(int r = 2; r <= 29; r++)
            {
                for(int c = 2; c <= 29; c++)
                {
                    unsigned char temp = 0;
                    file.read((char*) &temp, sizeof(temp));
                    // tp.at<uchar>(r, c) = (int) temp;
                    tp.at<float>(r, c) = (float) temp;
                    
                }
            }
            vec.push_back(tp);
        }
    }
}

void read_Mnist_Label(std::string filename, std::vector<int> &vec)
{
    std::ifstream file (filename, std::ios::binary);
    if (file.is_open())
    {
        int magic_number = 0;
        int number_of_images = 0;
        file.read((char*) &magic_number, sizeof(magic_number));
        magic_number = ReverseInt(magic_number);
        file.read((char*) &number_of_images,sizeof(number_of_images));
        number_of_images = ReverseInt(number_of_images);
        
        
        for(int i = 0; i < number_of_images; ++i)
        {
            unsigned char temp = 0;
            file.read((char*) &temp, sizeof(temp));
            vec.push_back((int) temp+1);
        }
    }
}


/* exampleMNISTSimilarity: Starting example for Manitest with a LibSVM classifier on the MNIST data */

bool file_exists(const char *fileName)
{
    std::ifstream infile(fileName);
    return infile.good();
}


void exampleMNIST(int argc, const char *argv[]) {
    
    int numArgs = 3;
    
    if (argc != numArgs) {
        std::cout << "Usage: exampleMNIST <PATH TO train-images-idx3-ubyte FILE> <PATH to LibSVM model>" << std::endl;
        exit(1);
    }
    
    if (!file_exists(argv[1]) || !file_exists(argv[2])) {
        std::cout << "File does not exist!" << std::endl;
        exit(1);
    }
    
    // Load the data
    std::vector<cv::Mat> trainImages;
    read_Mnist32(std::string(argv[1]), trainImages);
    
    Classifiers cl;
    
    // Create a LibSVM classifier object
    ClassifierLibSVM libsvm_classifier;
    // Load the model
    // libsvm_classifier.load("Classifiers/example_SVM_MNIST.model");
    libsvm_classifier.load(argv[2]);
    
    // Add to the list of classifiers we will be testing
    cl.addClassifier(&libsvm_classifier);
    
    // Add other classifiers as follows
    // MyClassifier classifierxyz;
    // classifierxyz.load("PATH TO MODEL");
    // cl.addClassifier(&classifierxyz)
    
    // NOTE: MyClassifier class should imperatively have a method int classify(Mat) that outputs the label of the Mat structure taken as input). Refer to ClassifierLibSVM.cpp for an example.
        
    // Create a transformation object (possible choices are currently Translation, ScalRot and Similarity).
    ScalRot trans;
    
    // Create a stencil object (describes the neighbouhood system of the discretized manifold)
    StencilCanonical st;
    
    st.hs = trans.getDefaultDiscretization();
    st.offsets = trans.getDefaultOffset();
    
    std::vector<double> avgScores(cl.numClassifiers());
    int numTrials = 1;
    
    for (int i=0;i<numTrials;i++) {
        std::cout << ">>> IMAGE #" << i << std::endl;
        cv::Mat img = trainImages[i];
        
        showImage("Original image " + std::to_string(i), img);

        // Save original image
        imwrite("orig_" + std::to_string(i) + ".png", img);
        
        // Convert image to double
        img.convertTo(img, CV_64F);
        
        // Initialize a Manitest object
        Manitest M = Manitest(img, (Transformation*) &trans, &st, &cl);
        
        // Compute the Manitest score
        clock_t begin = clock();
        M.run();
        clock_t end = clock();
        
        double elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
        std::cout << "Elapsed time = " << elapsed_secs << std::endl;
        
        // Show the Manitest images
        for (int j=0; j<M.scores.size(); j++) {
            std::ostringstream d; d << M.scores[j];
            showImage("Score = " + d.str(), M.images[j]);
            // Save transformed images
            imwrite("transformed_" + std::to_string(i) + "_" + std::to_string(j) + ".png", M.images[j]);
        }
        
        // Average Manitest scores
        for (int j=0; j<M.scores.size(); j++) {
            avgScores[j] = avgScores[j] + M.scores[j] / numTrials;
        }
        
    }
    
    std::cout << "Average scores = " << std::endl;
    for (int j=0; j<avgScores.size(); j++) {
        std::cout << avgScores[j] << " ";
    }
    
    std::cout << std::endl;
    
}